// TcpServ.cpp: implementation of the CTcpServ class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TcpServ.h"
#include "Interface.h"

#if _MSC_VER >= 1400						// using VS2005 platform or later
	#define IPv6_DUAL						// use IPv4 / IPv6 dual stack
#endif

#if defined IPv6_DUAL
	#include <ws2tcpip.h>
	#if _MSC_VER <= 1500					// using VS2008 platform or earlier
		#include <wspiapi.h>				// use getaddrinfo() wrapper for Win2k compatibility
	#endif
	#if !defined IPV6_V6ONLY
		#define IPV6_V6ONLY		27			// Vista specific definition
	#endif
#endif

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTcpServ::CTcpServ(CIL *pIL, UINT uPort, CWnd* pParent)
	: m_uPort(uPort)
	, CPilBoxSim(true)
{
	WSADATA wsd;

	m_pIL = pIL;
	m_pParent = pParent;

	m_nNumSocks = 0;
	m_cfd = INVALID_SOCKET;

	VERIFY(WSAStartup(MAKEWORD(2,2),&wsd) == 0);
}

CTcpServ::~CTcpServ()
{
	WSACleanup();
}

// --------------------------------
// StartServer()
// start the local TCP/IP server
// --------------------------------

bool CTcpServ::StartServer()
{
	m_nNumSocks = 0;

#if defined IPv6_DUAL
	// IPv4 / IPv6 implementation
	CHAR cPortIn[16];
	ADDRINFO sHints, *psAddrInfo, *psAI;

	SOCKET fd = 0;

	// the port no. as ASCII string
	sprintf_s(cPortIn,sizeof(cPortIn),"%u",m_uPort);

	memset(&sHints, 0, sizeof(sHints));
	sHints.ai_family = PF_UNSPEC;
	sHints.ai_socktype = SOCK_STREAM;
	sHints.ai_flags = AI_PASSIVE;

	if (getaddrinfo(NULL,cPortIn,&sHints,&psAddrInfo) != 0)
	{
		return false;
	}

	//
	// For each address getaddrinfo returned, we create a new socket,
	// bind that address to it, and create a queue to listen on.
	//
	for (m_nNumSocks = 0, psAI = psAddrInfo; psAI != NULL; psAI = psAI->ai_next)
	{
		if (m_nNumSocks == FD_SETSIZE)
		{
			break;
		}
		if ((psAI->ai_family != PF_INET) && (psAI->ai_family != PF_INET6))
		{
			// only PF_INET and PF_INET6 is supported
			continue;
		}
		fd = socket(psAI->ai_family,psAI->ai_socktype,psAI->ai_protocol);
		if (fd == INVALID_SOCKET)
		{
			// socket() failed
			continue;
		}
		if (psAI->ai_family == AF_INET6)
		{
			int ipv6only,optlen;

			if (   IN6_IS_ADDR_LINKLOCAL((PIN6_ADDR) &((PSOCKADDR_IN6) (psAI->ai_addr))->sin6_addr)
				&& (((PSOCKADDR_IN6) (psAI->ai_addr))->sin6_scope_id == 0)
			   )
			{
				// IPv6 link local addresses should specify a scope ID
				closesocket(fd);
				continue;
			}

			// this socket option is supported on Windows Vista or later
			optlen = sizeof(ipv6only);
			if (getsockopt(fd,IPPROTO_IPV6,IPV6_V6ONLY,(char*) &ipv6only,&optlen) == 0 && ipv6only == 0)
			{
				ipv6only = 1;				// set option

				// on Windows XP IPV6_V6ONLY is always set, on Vista and later set it manually
				if (setsockopt(fd,IPPROTO_IPV6,IPV6_V6ONLY,(char*) &ipv6only,sizeof(ipv6only)) == SOCKET_ERROR)
				{
					closesocket(fd);
					continue;
				}
			}
		}
		if (bind(fd,psAI->ai_addr,(int) psAI->ai_addrlen))
		{
			// bind() failed
			closesocket(fd);
			continue;
		}
		if (listen(fd,5) == SOCKET_ERROR)
		{
			// listen() failed
			closesocket(fd);
			continue;
		}
		m_sockfds[m_nNumSocks++] = fd;
		if (m_nNumSocks >= ARRAYSIZEOF(m_sockfds))
			break;
	}
	freeaddrinfo(psAddrInfo);

	if (m_nNumSocks == 0)					// socket not connected
	{
		return false;
	}
#else
	// IPv4 implementation
	struct sockaddr_in sServer;

	m_sockfds[m_nNumSocks] = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (m_sockfds[m_nNumSocks] == INVALID_SOCKET)
	{
		// socket() failed
		return false;
	}

	sServer.sin_family = AF_INET;
	sServer.sin_addr.s_addr = INADDR_ANY;
	sServer.sin_port = htons(m_uPort);

	if (bind(m_sockfds[m_nNumSocks], (LPSOCKADDR) &sServer, sizeof(sServer)) == SOCKET_ERROR)
	{
		// bind() failed
		closesocket(m_sockfds[m_nNumSocks]);
		return false;
	}

	if (listen(m_sockfds[m_nNumSocks], 5) == SOCKET_ERROR)
	{
		// listen() failed
		closesocket(m_sockfds[m_nNumSocks]);
		return false;
	}

	++m_nNumSocks;							// only 1 socket
#endif

	m_bRunning = true;
	AfxBeginThread(ThreadTcpIpServer,this);
	return true;
}

// --------------------------------
// StopServer()
// stop the local TCP/IP server
// --------------------------------

void CTcpServ::StopServer()
{
	INT i;

	m_bRunning = false; 					// delete server thread

	if (m_cfd != INVALID_SOCKET)
	{
		closesocket(m_cfd);
		m_cfd = INVALID_SOCKET;
	}

	for (i = 0; i < m_nNumSocks; ++i)
		closesocket(m_sockfds[i]);			// terminate select() for thread termination
	m_nNumSocks = 0;
	return;
}

// --------------------------------
// ThreadTcpIpServer
// local TCP/IP server thread
// --------------------------------

UINT CTcpServ::ThreadTcpIpServer(LPVOID pParam)
{
	CTcpServ *p = reinterpret_cast<CTcpServ *>(pParam);

	fd_set SockSet;
	int i;

	ASSERT(p->m_nNumSocks > 0);				// established server

	//
	// We now put the server into an external loop,
	// serving requests as they arrive.
	//
	FD_ZERO(&SockSet);
	while (p->m_bRunning)
	{
		p->m_cfd = INVALID_SOCKET;

		//
		// Check to see if we have any sockets remaining to be served
		// from previous time through this loop.  If not, call select()
		// to wait for a connection request or a datagram to arrive.
		//
		for (i = 0; i < p->m_nNumSocks; i++)
		{
			if (FD_ISSET(p->m_sockfds[i], &SockSet))
				break;
		}
		if (i == p->m_nNumSocks)			// all sockets
		{
			for (i = 0; i < p->m_nNumSocks; i++)
				FD_SET(p->m_sockfds[i], &SockSet);

			// select() can be finished by closesocket()
			if (select(p->m_nNumSocks, &SockSet, NULL, NULL, NULL) == SOCKET_ERROR)
			{
				// Win9x break with WSAEINTR (a blocking socket call was canceled)
				if (WSAEINTR != WSAGetLastError())
				{
					break;					// exit thread
				}
			}

			if (p->m_bRunning == false)		// exit thread
			{
				break;
			}
		}
		for (i = 0; i < p->m_nNumSocks; i++)
		{
			if (FD_ISSET(p->m_sockfds[i], &SockSet))
			{
				FD_CLR(p->m_sockfds[i], &SockSet);
				break;
			}
		}
		_ASSERT(i < p->m_nNumSocks);		// at least one socket is waiting
		if (i == p->m_nNumSocks) continue;	// no socket waiting

		//
		// Since this socket was returned by the select(), we know we
		// have a connection waiting and that this accept() won't block.
		//
		p->m_cfd = accept(p->m_sockfds[i], NULL, NULL);

		while (TRUE)
		{
			BYTE byData;

			INT nAmountRead = recv(p->m_cfd, (char *) &byData, sizeof(byData), 0);
			if (   nAmountRead == 0 		// client closed connection
				|| nAmountRead == SOCKET_ERROR)
			{
				closesocket(p->m_cfd);
				p->m_cfd = INVALID_SOCKET;
				break;
			}

			if (p->m_cfd == INVALID_SOCKET) break;

			byData = p->PILBox(byData);		// handle received data byte
			if (byData != 0)				// answer
			{
				// transmit
				INT nRet = send(p->m_cfd, (const char *) &byData, sizeof(byData), 0);
				if (nRet == SOCKET_ERROR)
				{
					closesocket(p->m_cfd);
					p->m_cfd = INVALID_SOCKET;
					break;
				}
			}
		}
	}

	p->m_pParent->PostMessage(WM_SERVERSTOPPED,reinterpret_cast<WPARAM>(p));
	AfxEndThread(0);
	return 0;
}
